from typing import Any, Callable, Iterable, List, Optional, TypeVar

import numpy as np
import pandas as pd
import pyarrow as pa
from triad.utils.pyarrow import apply_schema, to_pandas_dtype

T = TypeVar("T", bound=Any)


def as_array_iterable(
    df: T,
    schema: Optional[pa.Schema] = None,
    type_safe: bool = False,
    null_safe: bool = False,
) -> Iterable[List[Any]]:
    """Convert pandas like dataframe to iterable of rows in the format of list.

    :param df: pandas like dataframe
    :param schema: columns and types for output.
        Leave it None to return all columns in original type
    :param type_safe: whether to enforce the types in schema, if not, it will
        return the original values from the dataframe
    :param null_safe: whether to ensure returning null for nan or null values in
        columns with type int, bool and string
    :return: iterable of rows, each row is a list

    :Notice:
    * `null_safe` by default is False, for non pandas dataframe, setting it to
      True may cause errors
    * If there are nested types in schema, the conversion can be a lot slower
    """
    if df.shape[0] == 0:
        return
    if schema is None:
        schema = to_schema(df)
    else:
        df = df[schema.names]
        orig = to_schema(df)
        if not orig.equals(schema):
            df = enforce_type(df, schema, null_safe)
    if not type_safe or all(not pa.types.is_nested(x) for x in schema.types):
        for arr in df.itertuples(index=False, name=None):
            yield list(arr)
    else:
        # If schema has nested types, the conversion will be much slower
        for arr in apply_schema(
            schema,
            df.itertuples(index=False, name=None),
            copy=True,
            deep=True,
            str_as_json=True,
        ):
            yield arr


def to_schema(df: T) -> pa.Schema:
    """Extract pandas dataframe schema as pyarrow schema. This is a replacement
    of pyarrow.Schema.from_pandas, and it can correctly handle string type and
    empty dataframes

    :param df: pandas dataframe
    :raises ValueError: if pandas dataframe does not have named schema
    :return: pyarrow.Schema

    :Notice:
    The dataframe must be either empty, or with type pd.RangeIndex, pd.Int64Index
    or pd.UInt64Index and without a name, otherwise, `ValueError` will raise.
    """
    _ensure_compatible_index(df)
    if df.columns.dtype != "object":
        raise ValueError("Pandas dataframe must have named schema")
    if isinstance(df, pd.DataFrame) and len(df.index) > 0:
        return pa.Schema.from_pandas(df)
    fields: List[pa.Field] = []
    for i in range(df.shape[1]):
        tp = df.dtypes[i]
        if tp == np.dtype("object") or tp == np.dtype(str):
            t = pa.string()
        else:
            t = pa.from_numpy_dtype(tp)
        fields.append(pa.field(df.columns[i], t))
    return pa.schema(fields)


def enforce_type(df: T, schema: pa.Schema, null_safe: bool = False) -> T:
    """Enforce the pandas like dataframe to comply with `schema`.

    :param df: pandas like dataframe
    :param schema: pyarrow schema
    :param null_safe: whether to enforce None value for int, string and bool values
    :return: converted dataframe

    :Notice:
    When `null_safe` is true, the native column types in the dataframe may change,
    for example, if a column of `int64` has None values, the output will make sure
    each value in the column is either None or an integer, however, due to the
    behavior of pandas like dataframes, the type of the columns may
    no longer be `int64`
    """
    if len(df.index) == 0:
        return df
    if not null_safe:
        return df.astype(dtype=to_pandas_dtype(schema))
    for v in schema:
        s = df[v.name]
        if pa.types.is_string(v.type):
            ns = s[s.isnull()].index.tolist()
            s = s.astype(str)
            s.iloc[ns] = None
        elif pa.types.is_integer(v.type) or pa.types.is_boolean(v.type):
            ns = s[s.isnull()].index.tolist()
            s = s.fillna(0).astype(v.type.to_pandas_dtype())
            s.iloc[ns] = None
        else:
            s = s.astype(v.type.to_pandas_dtype())
        df[v.name] = s
    return df


def safe_groupby_apply(
    df: Any,
    cols: List[str],
    func: Callable[[T], T],
    key_col_name="__safe_groupby_key__",
) -> T:
    """Safe groupby apply operation on pandas like dataframes

    :param df: pandas like dataframe
    :param cols: columns to group on, can be empty
    :param func: apply function, df in, df out
    :param key_col_name: temp key as index for groupu. default "__safe_groupby_key__"
    :return: output dataframe

    :Notice:
    The dataframe must be either empty, or with type pd.RangeIndex, pd.Int64Index
    or pd.UInt64Index and without a name, otherwise, `ValueError` will raise.
    """
    _ensure_compatible_index(df)
    if len(cols) == 0:
        return func(df)
    keys = df[cols].drop_duplicates()
    keys[key_col_name] = np.arange(keys.shape[0])
    df = df.merge(keys, on=cols).set_index([key_col_name])

    def _wrapper(df: T) -> T:
        return func(df.reset_index(drop=True))

    return df.groupby([key_col_name]).apply(_wrapper).reset_index(drop=True)


def _ensure_compatible_index(df: T) -> None:
    if df.index.name is not None:
        raise ValueError(f"pandas like datafame index can't have name")
    if isinstance(df.index, (pd.RangeIndex, pd.Int64Index, pd.UInt64Index)):
        return
    if len(df.index) == 0:
        return
    raise ValueError(
        f"pandas like datafame must have default index, but got {type(df.index)}"
    )



# flake8: noqa
from triad.collections.dataframe.array_dataframe import ArrayDataFrame
from triad.collections.dataframe.dataframe import DataFrame, LocalDataFrame
from triad.collections.dataframe.iterable_dataframe import IterableDataFrame
from triad.collections.dataframe.pandas_dataframes import PandasDataFrame



import json
from abc import ABC, abstractmethod
from threading import RLock
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union

import pandas as pd
from triad.collections.dict import ParamDict
from triad.collections.schema import Schema


class DataFrame(ABC):
    SHOW_LOCK = RLock()

    def __init__(self, schema: Any = None, metadata: Any = None):
        if not callable(schema):
            schema = Schema(schema).assert_not_empty()
            schema.set_readonly()
            self._schema: Union[Schema, Callable[[], Schema]] = schema
            self._schema_discovered = True
        else:
            self._schema: Union[Schema, Callable[[], Schema]] = schema  # type: ignore
            self._schema_discovered = False
        self._metadata = ParamDict(metadata, deep=True)
        self._lazy_schema_lock = RLock()

    @property
    def metadata(self) -> ParamDict:
        return self._metadata

    @property
    def schema(self) -> Schema:
        if self._schema_discovered:
            # we must keep it simple because it could be called on every row by a user
            assert isinstance(self._schema, Schema)
            return self._schema  # type: ignore
        with self._lazy_schema_lock:
            self._schema = Schema(self._schema()).assert_not_empty()  # type: ignore
            self._schema.set_readonly()
            self._schema_discovered = True
            return self._schema

    @property
    @abstractmethod
    def is_local(self) -> bool:  # pragma: no cover
        raise NotImplementedError

    # @abstractmethod
    # def as_local(self) -> "DataFrame":  # pragma: no cover
    #    raise NotImplementedError

    # @abstractmethod
    # def apply_schema(self, schema: Any) -> None:  # pragma: no cover
    #    raise NotImplementedError

    # @abstractmethod
    # def num_partitions(self) -> int:  # pragma: no cover
    #    raise NotImplementedError

    @abstractmethod
    def empty(self) -> bool:  # pragma: no cover
        raise NotImplementedError

    @abstractmethod
    def peek_array(self) -> Any:  # pragma: no cover
        raise NotImplementedError

    def peek_dict(self) -> Dict[str, Any]:
        arr = self.peek_array()
        return {self.schema.names[i]: arr[i] for i in range(len(self.schema))}

    @abstractmethod
    def count(self, persist: bool = False) -> int:  # pragma: no cover
        raise NotImplementedError

    def as_pandas(self) -> pd.DataFrame:
        pdf = pd.DataFrame(self.as_array(), columns=self.schema.names)
        return pdf.astype(dtype=self.schema.pd_dtype)

    # @abstractmethod
    # def as_pyarrow(self) -> pa.Table:  # pragma: no cover
    #    raise NotImplementedError

    @abstractmethod
    def as_array(
        self, columns: Optional[List[str]] = None, type_safe: bool = False
    ) -> List[Any]:  # pragma: no cover
        raise NotImplementedError

    @abstractmethod
    def as_array_iterable(
        self, columns: Optional[List[str]] = None, type_safe: bool = False
    ) -> Iterable[Any]:  # pragma: no cover
        raise NotImplementedError

    @abstractmethod
    def drop(self, cols: List[str]) -> "DataFrame":  # pragma: no cover
        raise NotImplementedError

    def show(
        self,
        n: int = 10,
        show_count: bool = False,
        title: Optional[str] = None,
        best_width: int = 100,
    ) -> None:
        arr: List[List[str]] = self.head(n)
        count = -1
        if len(arr) < n:
            count = len(arr)
        elif show_count:
            count = self.count()
        with DataFrame.SHOW_LOCK:
            if title is not None:
                print(title)
            print(type(self).__name__)
            tb = _PrettyTable(self.schema, arr, best_width)
            print("\n".join(tb.to_string()))
            if count >= 0:
                print(f"Total count: {count}")
                print("")
            if len(self.metadata) > 0:
                print("Metadata:")
                print(self.metadata.to_json(indent=True))
                print("")

    def head(self, n: int, columns: Optional[List[str]] = None) -> List[Any]:
        res: List[Any] = []
        for row in self.as_array_iterable(columns, type_safe=True):
            if n < 1:
                break
            res.append(list(row))
            n -= 1
        return res

    def as_dict_iterable(
        self, columns: Optional[List[str]] = None
    ) -> Iterable[Dict[str, Any]]:
        if columns is None:
            columns = self.schema.names
        idx = range(len(columns))
        for x in self.as_array_iterable(columns, type_safe=True):
            yield {columns[i]: x[i] for i in idx}

    def get_info_str(self) -> str:
        return json.dumps(
            {
                "schema": str(self.schema),
                "type": "{}.{}".format(
                    self.__class__.__module__, self.__class__.__name__
                ),
                "metadata": self.metadata,
            }
        )


class LocalDataFrame(DataFrame):
    def __init__(self, schema: Any = None, metadata: Any = None):
        super().__init__(schema=schema, metadata=metadata)

    @property
    def is_local(self):
        return True


class _PrettyTable(object):
    def __init__(
        self,  # noqa: C901
        schema: Schema,
        data: List[Any],
        best_width: int,
        truncate_width: int = 500,
    ):
        raw: List[List[str]] = []
        self.titles = str(schema).split(",")
        col_width_min = [len(t) for t in self.titles]
        col_width_max = list(col_width_min)
        self.col_width = list(col_width_min)
        # Convert all cells to string with truncation
        for row in data:
            raw_row: List[str] = []
            for i in range(len(schema)):
                d = self._cell_to_raw_str(row[i], truncate_width)
                col_width_max[i] = max(col_width_max[i], len(d))
                raw_row.append(d)
            raw.append(raw_row)
        # Adjust col width based on best_width
        # It find the remaining width after fill all cols with min widths,
        # and redistribute the remaining based on the diff between max and min widths
        dt = sorted(
            filter(  # noqa: C407
                lambda x: x[0] > 0,
                [(w - col_width_min[i], i) for i, w in enumerate(col_width_max)],
            )
        )
        if len(dt) > 0:
            remaining = max(0, best_width - sum(col_width_min) - len(col_width_min)) + 1
            total = sum(x[0] for x in dt)
            for diff, index in dt:
                if remaining <= 0:  # pragma: no cover
                    break
                given = remaining * diff // total
                remaining -= given
                self.col_width[index] += given
        # construct data -> List[List[List[str]]], make sure on the same row, each cell
        # has the same length of strings
        self.data = [
            [self._wrap(row[i], self.col_width[i]) for i in range(len(schema))]
            for row in raw
        ]
        blank = ["".ljust(self.col_width[i]) for i in range(len(schema))]
        for row in self.data:
            max_h = max(len(c) for c in row)
            for i in range(len(schema)):
                row[i] += [blank[i]] * (max_h - len(row[i]))

    def to_string(self) -> Iterable[str]:
        yield "|".join(
            self.titles[i].ljust(self.col_width[i]) for i in range(len(self.titles))
        )
        yield "+".join(
            "".ljust(self.col_width[i], "-") for i in range(len(self.titles))
        )
        for row in self.data:
            for tp in zip(*row):
                yield "|".join(tp)

    def _cell_to_raw_str(self, obj: Any, truncate_width: int) -> str:
        raw = "NULL" if obj is None else str(obj)
        if len(raw) > truncate_width:
            raw = raw[: max(0, truncate_width - 3)] + "..."
        return raw

    def _wrap(self, s: str, width: int) -> List[str]:
        res: List[str] = []
        start = 0
        while start < len(s):
            end = min(len(s), start + width)
            sub = s[start:end]
            if end < len(s):
                res.append(sub)
            else:
                res.append(sub.ljust(width, " "))
                break
            start += width
        return res


def _get_schema_change(
    orig_schema: Optional[Schema], schema: Any
) -> Tuple[Schema, List[int]]:
    if orig_schema is None:
        schema = Schema(schema).assert_not_empty()
        return schema, []
    elif schema is None:
        return orig_schema.assert_not_empty(), []
    if isinstance(schema, (str, Schema)) and orig_schema == schema:
        return orig_schema.assert_not_empty(), []
    if schema in orig_schema:
        # keys list or schema like object that is a subset of orig
        schema = orig_schema.extract(schema).assert_not_empty()
        pos = [orig_schema.index_of_key(x) for x in schema.names]
        if pos == list(range(len(orig_schema))):
            pos = []
        return schema, pos
    # otherwise it has to be a schema like object that must be a subset
    # of orig, and that has mismatched types
    schema = Schema(schema).assert_not_empty()
    pos = [orig_schema.index_of_key(x) for x in schema.names]
    if pos == list(range(len(orig_schema))):
        pos = []
    return schema, pos



from typing import Any, Iterable, List, Optional, Tuple

import pandas as pd
import pyarrow as pa
from triad.collections.dataframe.dataframe import DataFrame, LocalDataFrame
from triad.collections.schema import Schema
from triad.exceptions import InvalidOperationError
from triad.utils.assertion import assert_arg_not_none, assert_or_throw
from triad.utils.pyarrow import apply_schema


class PandasDataFrame(LocalDataFrame):
    def __init__(  # noqa: C901
        self, df: Any = None, schema: Any = None, metadata: Any = None
    ):
        if df is None:
            schema = Schema(schema).assert_not_empty()
            pdf = pd.DataFrame([], columns=schema.names)
            pdf = pdf.astype(dtype=schema.pd_dtype)
        elif isinstance(df, PandasDataFrame):
            pdf = df.native
            schema = None
        elif isinstance(df, (pd.DataFrame, pd.Series)):
            if isinstance(df, pd.Series):
                df = df.to_frame()
            pdf = df
            schema = None if schema is None else Schema(schema)
        elif isinstance(df, Iterable):
            assert_arg_not_none(schema, msg=f"schema can't be None for iterable input")
            schema = Schema(schema).assert_not_empty()
            pdf = pd.DataFrame(df, columns=schema.names)
            pdf = pdf.astype(dtype=schema.pd_dtype)
        else:
            raise ValueError(f"{df} is incompatible with PandasDataFrame")
        pdf, schema = self._apply_schema(pdf, schema)
        super().__init__(schema, metadata)
        self._native = pdf

    @property
    def native(self) -> pd.DataFrame:
        return self._native

    def empty(self) -> bool:
        return self.native.empty

    def peek_array(self) -> Any:
        return self.native.iloc[0].values.tolist()

    def count(self, persist: bool = False) -> int:
        return self.native.shape[0]

    def as_pandas(self) -> pd.DataFrame:
        return self._native

    def drop(self, cols: List[str]) -> DataFrame:
        try:
            schema = self.schema - cols
        except Exception as e:
            raise InvalidOperationError(str(e))
        if len(schema) == 0:
            raise InvalidOperationError("Can't remove all columns of a dataframe")
        return PandasDataFrame(self.native.drop(cols, axis=1), schema)

    def as_array(
        self, columns: Optional[List[str]] = None, type_safe: bool = False
    ) -> List[Any]:
        return list(self.as_array_iterable(columns, type_safe=type_safe))

    def as_array_iterable(
        self, columns: Optional[List[str]] = None, type_safe: bool = False
    ) -> Iterable[Any]:
        if self._native.shape[0] == 0:
            return
        sub = self.schema if columns is None else self.schema.extract(columns)
        df = self._native[sub.names]
        if not type_safe or all(
            not isinstance(x, (pa.StructType, pa.ListType)) for x in self.schema.types
        ):
            for arr in df.itertuples(index=False, name=None):
                yield list(arr)
        else:  # TODO: If schema has nested types, the conversion will be much slower
            for arr in apply_schema(
                self.schema.pa_schema,
                df.itertuples(index=False, name=None),
                copy=True,
                deep=True,
                str_as_json=True,
            ):
                yield arr

    def _apply_schema(
        self, pdf: pd.DataFrame, schema: Optional[Schema]
    ) -> Tuple[pd.DataFrame, Schema]:
        assert_or_throw(
            pdf.empty or type(pdf.index) == pd.RangeIndex,
            ValueError("Pandas datafame must have default index"),
        )
        if pdf.columns.dtype == "object":  # pdf has named schema
            pschema = Schema(pdf)
            if schema is None or pschema == schema:
                return pdf, pschema.assert_not_empty()
            pdf = pdf[schema.assert_not_empty().names]
        else:  # pdf has no named schema
            schema = Schema(schema).assert_not_empty()
            assert_or_throw(
                pdf.shape[1] == len(schema),
                ValueError(f"Pandas datafame column count doesn't match {schema}"),
            )
            pdf.columns = schema.names
        pdf = pdf.astype(dtype=schema.pd_dtype)
        return pdf, schema



from typing import Any, Iterable, List, Optional

from triad.collections.dataframe.dataframe import (
    DataFrame,
    LocalDataFrame,
    _get_schema_change,
)
from triad.exceptions import InvalidOperationError
from triad.utils.assertion import assert_or_throw
from triad.utils.pyarrow import apply_schema


class ArrayDataFrame(LocalDataFrame):
    def __init__(  # noqa: C901
        self, df: Any = None, schema: Any = None, metadata: Any = None
    ):
        if df is None:
            super().__init__(schema, metadata)
            self._native = []
        elif isinstance(df, DataFrame):
            if schema is None:
                super().__init__(df.schema, metadata)
                self._native = df.as_array(type_safe=False)
            else:
                schema, _ = _get_schema_change(df.schema, schema)
                super().__init__(schema, metadata)
                self._native = df.as_array(schema.names, type_safe=False)
        elif isinstance(df, Iterable):
            super().__init__(schema, metadata)
            self._native = df if isinstance(df, List) else list(df)
        else:
            raise ValueError(f"{df} is incompatible with ArrayDataFrame")

    @property
    def native(self) -> List[Any]:
        return self._native

    def empty(self) -> bool:
        return self.count() == 0

    def peek_array(self) -> Any:
        return list(self.native[0])

    def count(self, persist: bool = False) -> int:
        return len(self.native)

    def drop(self, cols: List[str]) -> DataFrame:
        try:
            schema = self.schema - cols
        except Exception as e:
            raise InvalidOperationError(str(e))
        if len(schema) == 0:
            raise InvalidOperationError("Can't remove all columns of a dataframe")
        return ArrayDataFrame(self, schema)

    def as_array(
        self, columns: Optional[List[str]] = None, type_safe: bool = False
    ) -> List[Any]:
        if not type_safe and columns is None:
            return self.native
        return list(self.as_array_iterable(columns, type_safe=type_safe))

    def as_array_iterable(
        self, columns: Optional[List[str]] = None, type_safe: bool = False
    ) -> Iterable[Any]:
        if columns is None:
            pos = []
        else:
            pos = [self.schema.index_of_key(k) for k in columns]
            assert_or_throw(len(pos) > 0, "columns if set must be non empty")
        if not type_safe:
            for item in self._iter_cols(pos):
                yield item
        else:
            sub = self.schema if columns is None else self.schema.extract(columns)
            for item in apply_schema(
                sub.pa_schema,
                self._iter_cols(pos),
                copy=True,
                deep=True,
                str_as_json=True,
            ):
                yield item

    def _iter_cols(self, pos: List[int]) -> Iterable[List[Any]]:
        if len(pos) == 0:
            for row in self.native:
                yield row
        else:
            for row in self.native:
                yield [row[p] for p in pos]



from collections import OrderedDict
from typing import Any, Dict, List, Optional, Set, Tuple

import numpy as np
import pandas as pd
import pyarrow as pa
from triad.collections.dict import IndexedOrderedDict
from triad.utils.assertion import assert_arg_not_none, assert_or_throw
from triad.utils.pyarrow import (
    expression_to_schema,
    is_supported,
    schema_to_expression,
    to_pa_datatype,
    to_pandas_dtype,
    validate_column_name,
)
from triad.utils.pandas_like import to_schema


class SchemaError(Exception):
    """Exceptions related with construction and modifying schemas
    """

    def __init__(self, message: str):
        super().__init__(message)


class Schema(IndexedOrderedDict[str, pa.Field]):
    """A Schema wrapper on top of pyarrow.Fields. This has more features
    than pyarrow.Schema, and they can convert to each other.

    This class can be initialized from schema like objects. Here is a list of
    schema like objects:
    * pyarrow.Schema or Schema objects
    * pyarrow.Field: single field will be treated as a single column schema
    * schema expressions: :func:`~triad.utils.pyarrow.expression_to_schema`
    * Dict[str,Any]: key will be the columns, and value will be type like objects
    * Tuple[str,Any]: first item will be the only column name of the schema,
        and the second has to be a type like object
    * List[Any]: a list of Schema like objects
    * pandas.DataFrame: it will extract the dataframe's schema

    Here is a list of data type like objects:
    * pyarrow.DataType
    * pyarrow.Field: will only use the type attribute of the field
    * type expression or other objects: for :func:`~triad.utils.pyarrow.to_pa_datatype`

    The column name must follow the rules
    by :func:`~triad.utils.pyarrow.validate_column_name`

    :Examples:
    >>> Schema("a:int,b:int")
    >>> Schema("a:int","b:int")
    >>> Schema(a=int,b=str) # == Schema("a:long,b:str")
    >>> Schema(dict(a=int,b=str)) # == Schema("a:long,b:str")
    >>> Schema([(a,int),(b,str)]) # == Schema("a:long,b:str")
    >>> Schema((a,int),(b,str)) # == Schema("a:long,b:str")
    >>> Schema("a:[int],b:{x:int,y:{z:[str],w:byte}},c:[{x:str}]")

    :Notice:
    * For supported pyarrow.DataTypes see :func:`~triad.utils.pyarrow.is_supported`
    * If you use python type as data type (e.g. `Schema(a=int,b=str)`) be aware
        the data type different. (e.g. python `int` type -> pyarrow `long`/`int64` type)
    * When not readonly, only `append` is allowed, `update` or `remove` are disallowed
    * When readonly, no modification on the existing schema is allowed
    * `append`, `update` and `remove` are always allowed when creating a new object
    * InvalidOperationError will be raised for disallowed operations
    * At most one of `*args` and `**kwargs` can be set

    :param args: one or multiple schema like objects, which will be combined in order
    :param kwargs: key value pairs for the schema
    """

    def __init__(self, *args: Any, **kwargs: Any):
        if len(args) > 0 and len(kwargs) > 0:
            raise SchemaError("Can't set both *args and **kwargs")
        if len(args) == 1:  # duplicate code for better performance
            if isinstance(args[0], Schema):
                super().__init__(args[0])  # type: ignore
                return
            fields: Optional[List[pa.Field]] = None
            if isinstance(args[0], str):
                fields = list(expression_to_schema(args[0]))
            if isinstance(args[0], pa.Schema):
                fields = list(args[0])
            if isinstance(args[0], pa.Field):
                fields = [args[0]]
            if fields is not None:
                fields = [self._validate_field(f) for f in fields]
                super().__init__([(x.name, x) for x in fields])
                return
        super().__init__()
        if len(args) > 0:
            self.append(list(args))
        elif len(kwargs) > 0:
            self.append(kwargs)

    @property
    def names(self) -> List[str]:
        """List of column names
        """
        self._build_index()
        return self._index_key  # type: ignore

    @property
    def fields(self) -> List[pa.Field]:
        """List of pyarrow.Fields
        """
        return list(self.values())

    @property
    def types(self) -> List[pa.DataType]:
        """List of pyarrow.DataTypes
        """
        return [v.type for v in self.values()]

    @property
    def pyarrow_schema(self) -> pa.Schema:
        """convert as pyarrow.Schema
        """
        return pa.schema(self.fields)

    @property
    def pa_schema(self) -> pa.Schema:
        """convert as pyarrow.Schema
        """
        return self.pyarrow_schema

    @property
    def pandas_dtype(self) -> Dict[str, np.dtype]:
        """convert as `dtype` dict for pandas dataframes.
        Currently, struct type is not supported
        """
        return to_pandas_dtype(self.pa_schema)

    @property
    def pd_dtype(self) -> Dict[str, np.dtype]:
        """convert as `dtype` dict for pandas dataframes.
        Currently, struct type is not supported
        """
        return self.pandas_dtype

    def assert_not_empty(self) -> "Schema":
        if len(self) > 0:
            return self
        raise SchemaError("Schema can't be empty")

    def copy(self) -> "Schema":
        """Clone Schema object

        :return: cloned object
        """
        other = super().copy()
        assert isinstance(other, Schema)
        return other

    def __repr__(self) -> str:
        return schema_to_expression(self.pyarrow_schema)

    def __iadd__(self, obj: Any) -> "Schema":
        return self.append(obj)

    def __isub__(self, obj: Any) -> "Schema":
        raise SchemaError("'-=' is not allowed for Schema")

    def __add__(self, obj: Any) -> "Schema":
        return self.copy().append(obj)

    def __sub__(self, obj: Any) -> "Schema":
        return self.remove(
            obj,
            ignore_key_mismatch=False,
            require_type_match=True,
            ignore_type_mismatch=False,
        )

    def __setitem__(  # type: ignore
        self, name: str, value: Any, *args: List[Any], **kwds: Dict[str, Any]
    ) -> None:
        assert_arg_not_none(value, "value")
        if not validate_column_name(name):
            raise SchemaError(f"Invalid column name {name}")
        if name in self:  # update existing value is not allowed
            raise SchemaError(f"{name} already exists in {self}")
        if isinstance(value, pa.Field):
            assert_or_throw(
                name == value.name, SchemaError(f"{name} doesn't match {value}")
            )
        elif isinstance(value, pa.DataType):
            value = pa.field(name, value)
        else:
            value = pa.field(name, to_pa_datatype(value))
        assert_or_throw(
            is_supported(value.type), SchemaError(f"{value} is not supported")
        )
        super().__setitem__(name, value, *args, **kwds)  # type: ignore

    def __eq__(self, other: Any) -> bool:
        if other is None:
            return False
        if isinstance(other, Schema):
            return super().__eq__(other)
        if isinstance(other, str):
            return self.__repr__() == other
        return self == Schema(other)

    def __contains__(self, key: Any) -> bool:  # noqa: C901
        if key is None:
            return False
        if isinstance(key, str):
            if ":" not in key:
                return super().__contains__(key)
        elif isinstance(key, pa.Field):
            res = super().get(key.name, None)
            if res is None:
                return False
            return key.type == res.type
        elif isinstance(key, Schema):
            for f in key.values():
                if f not in self:
                    return False
            return True
        elif isinstance(key, List):
            for f in key:
                if f not in self:
                    return False
            return True
        return Schema(key) in self

    def append(self, obj: Any) -> "Schema":  # noqa: C901
        """Append schema like object to the current schema. Only new columns
        are allowed.

        :raises SchemaError: if a column exists or is invalid or obj is not convertible
        :return: the Schema object itself
        """
        try:
            if obj is None:
                return self
            elif isinstance(obj, pa.Field):
                self[obj.name] = obj.type
            elif isinstance(obj, str):
                self._append_pa_schema(expression_to_schema(obj))
            elif isinstance(obj, Dict):
                for k, v in obj.items():
                    self[k] = v
            elif isinstance(obj, pa.Schema):
                self._append_pa_schema(obj)
            elif isinstance(obj, pd.DataFrame):
                self._append_pa_schema(to_schema(obj))
            elif isinstance(obj, Tuple):  # type: ignore
                self[obj[0]] = obj[1]
            elif isinstance(obj, List):
                for x in obj:
                    self.append(x)
            else:
                raise SchemaError(f"Invalid schema to add {obj}")
            return self
        except SchemaError:
            raise
        except Exception as e:
            raise SchemaError(str(e))

    def remove(  # noqa: C901
        self,
        obj: Any,
        ignore_key_mismatch: bool = False,
        require_type_match: bool = True,
        ignore_type_mismatch: bool = False,
    ) -> "Schema":
        if obj is None:
            return self.copy()
        target = self
        if isinstance(obj, str):
            if ":" in obj:  # expression
                ps = expression_to_schema(obj)
                pairs: List[Tuple[str, pa.DataType]] = list(zip(ps.names, ps.types))
            else:
                pairs = [(obj, None)]  # single key
        elif isinstance(obj, (pa.Schema, Schema)):
            pairs = list(zip(obj.names, obj.types))
        elif isinstance(obj, (List, Set)):
            keys: List[str] = []
            other: List[Any] = []
            for x in obj:
                if isinstance(x, str) and ":" not in x:
                    keys.append(x)
                else:
                    other.append(x)
            pairs = [(x, None) for x in keys]
            for o in other:
                target = target.remove(
                    o,
                    ignore_key_mismatch=ignore_key_mismatch,
                    require_type_match=require_type_match,
                    ignore_type_mismatch=ignore_type_mismatch,
                )
        else:
            return self.remove(
                Schema(obj),
                ignore_key_mismatch=ignore_key_mismatch,
                require_type_match=require_type_match,
                ignore_type_mismatch=ignore_type_mismatch,
            )
        od = OrderedDict(target)
        for k, v in pairs:
            k = k.strip()
            if k == "":
                continue
            if k not in od:
                if ignore_key_mismatch:
                    continue
                raise SchemaError(f"Can't remove {k} from {target}")
            if v is None:
                del od[k]
            else:
                tp = od[k].type
                if not require_type_match or tp == v:
                    del od[k]
                elif not ignore_type_mismatch:
                    raise SchemaError(
                        f"Unable to remove {k}:{v} from {self}, type mismatch"
                    )
        return Schema(od)

    def extract(  # noqa: C901
        self,
        obj: Any,
        ignore_key_mismatch: bool = False,
        require_type_match: bool = True,
        ignore_type_mismatch: bool = False,
    ) -> "Schema":
        if obj is None:
            return Schema()
        if isinstance(obj, str):
            if ":" in obj:  # expression
                ps = expression_to_schema(obj)
                pairs: List[Tuple[str, pa.DataType]] = list(zip(ps.names, ps.types))
            else:
                pairs = [(obj, None)]  # single key
        elif isinstance(obj, (pa.Schema, Schema)):
            pairs = list(zip(obj.names, obj.types))
        elif isinstance(obj, List):
            fields: List[pa.Field] = []
            for x in obj:
                if isinstance(x, str) and ":" not in x:
                    if x not in self:
                        if not ignore_key_mismatch:
                            raise SchemaError(f"Can't extract {x} from {self}")
                    else:
                        fields.append(self[x])
                else:
                    fields += self.extract(
                        x,
                        ignore_key_mismatch=ignore_key_mismatch,
                        require_type_match=require_type_match,
                        ignore_type_mismatch=ignore_type_mismatch,
                    ).fields
            return Schema(pa.schema(fields))
        else:
            return self.extract(
                Schema(obj),
                ignore_key_mismatch=ignore_key_mismatch,
                require_type_match=require_type_match,
                ignore_type_mismatch=ignore_type_mismatch,
            )
        fields = []
        for k, v in pairs:
            k = k.strip()
            if k == "":
                continue
            if k not in self:
                if ignore_key_mismatch:
                    continue
                raise SchemaError(f"Can't extract {k} from {self}")
            if v is None:
                fields.append(self[k])
            else:
                tp = self[k].type
                if not require_type_match or tp == v:
                    fields.append(self[k])
                elif not ignore_type_mismatch:
                    raise SchemaError(
                        f"Unable to extract {k}:{v} from {self}, type mismatch"
                    )
        return Schema(pa.schema(fields))

    def exclude(
        self,
        other: Any,
        require_type_match: bool = True,
        ignore_type_mismatch: bool = False,
    ) -> "Schema":
        return self.remove(
            other,
            ignore_key_mismatch=True,
            require_type_match=require_type_match,
            ignore_type_mismatch=ignore_type_mismatch,
        )

    def intersect(
        self,
        other: Any,
        require_type_match: bool = True,
        ignore_type_mismatch: bool = True,
        use_other_order: bool = False,
    ) -> "Schema":
        if not use_other_order:
            diff = self.exclude(
                other,
                require_type_match=require_type_match,
                ignore_type_mismatch=ignore_type_mismatch,
            )
            return self - diff
        else:
            return self.extract(
                other,
                require_type_match=require_type_match,
                ignore_type_mismatch=ignore_type_mismatch,
            )

    def union(self, other: Any, require_type_match: bool = False) -> "Schema":
        return self.copy().union_with(other, require_type_match=require_type_match)

    def union_with(self, other: Any, require_type_match: bool = False) -> "Schema":
        if not isinstance(other, Schema):
            other = Schema(other)
        try:
            other = other.exclude(
                self, require_type_match=require_type_match, ignore_type_mismatch=False
            )
            self += other
            return self
        except Exception as e:
            raise SchemaError(f"Unable to union {self} with {other}: {str(e)}")

    def rename(self, columns: Dict[str, str], ignore_missing: bool = False) -> "Schema":
        """Rename the current schema and generate a new one

        :param columns: dictionary to map from old to new column names
        :return: renamed schema object
        """
        if not ignore_missing:
            for x in columns:
                if x not in self:
                    raise SchemaError(f"Failed to rename: {x} not in {self}")
        pairs = [
            (k if k not in columns else columns[k], v.type) for k, v in self.items()
        ]
        return Schema(pairs)

    def _pre_update(self, op: str, need_reindex: bool = True) -> None:
        if op == "__setitem__":
            super()._pre_update(op, need_reindex)
        else:
            raise SchemaError(f"{op} is not allowed in Schema")

    def _append_pa_schema(self, other: pa.Schema) -> "Schema":
        for k, v in zip(other.names, other.types):
            self[k] = v
        return self

    def _validate_field(self, field: pa.Field) -> pa.Field:
        assert_or_throw(
            validate_column_name(field.name), SchemaError(f"{field} name is invalid")
        )
        assert_or_throw(
            is_supported(field.type), SchemaError(f"{field} type is not supported")
        )
        return field



import json
import math
from datetime import datetime

import numpy as np
import pandas as pd
import pyarrow as pa
from pytest import raises
from triad.utils.pandas_like import (_ensure_compatible_index,
                                     as_array_iterable, enforce_type,
                                     safe_groupby_apply, to_schema)
from triad.utils.pyarrow import expression_to_schema


def test_to_schema():
    df = pd.DataFrame([[1.0, 2], [2.0, 3]])
    raises(ValueError, lambda: to_schema(df))
    df = pd.DataFrame([[1.0, 2], [2.0, 3]], columns=["x", "y"])
    assert list(pa.Schema.from_pandas(df)) == list(to_schema(df))
    df = pd.DataFrame([["a", 2], ["b", 3]], columns=["x", "y"])
    assert list(pa.Schema.from_pandas(df)) == list(to_schema(df))
    df = pd.DataFrame([], columns=["x", "y"])
    df = df.astype(dtype={"x": np.int32, "y": np.dtype('object')})
    assert [pa.field("x", pa.int32()), pa.field(
        "y", pa.string())] == list(to_schema(df))
    df = pd.DataFrame([[1, "x"], [2, "y"]], columns=["x", "y"])
    df = df.astype(dtype={"x": np.int32, "y": np.dtype('object')})
    assert list(pa.Schema.from_pandas(df)) == list(to_schema(df))
    df = pd.DataFrame([[1, "x"], [2, "y"]], columns=["x", "y"])
    df = df.astype(dtype={"x": np.int32, "y": np.dtype(str)})
    assert list(pa.Schema.from_pandas(df)) == list(to_schema(df))
    df = pd.DataFrame([[1, "x"], [2, "y"]], columns=["x", "y"])
    df = df.astype(dtype={"x": np.int32, "y": np.dtype('str')})
    assert list(pa.Schema.from_pandas(df)) == list(to_schema(df))

    # test index
    df = pd.DataFrame([[3.0, 2], [2.0, 3]], columns=["x", "y"])
    df = df.sort_values(["x"])
    assert list(pa.Schema.from_pandas(df)) == list(to_schema(df))
    df.index.name = "x"
    raises(ValueError, lambda: to_schema(df))
    df = df.reset_index(drop=True)
    assert list(pa.Schema.from_pandas(df)) == list(to_schema(df))
    df["p"] = "p"
    df = df.set_index(["p"])
    df.index.name = None
    raises(ValueError, lambda: to_schema(df))


def test_as_array_iterable():
    df = DF([], "a:str,b:int")
    assert [] == df.as_array()
    assert [] == df.as_array(type_safe=True)

    df = DF([["a", 1]], "a:str,b:int")
    assert [["a", 1]] == df.as_array()
    assert [["a", 1]] == df.as_array(["a", "b"])
    assert [[1, "a"]] == df.as_array(["b", "a"])

    # prevent pandas auto type casting
    df = DF([[1.0, 1.1]], "a:double,b:int")
    assert [[1.0, 1]] == df.as_array()
    assert isinstance(df.as_array()[0][0], float)
    assert isinstance(df.as_array()[0][1], int)
    assert [[1.0, 1]] == df.as_array(["a", "b"])
    assert [[1, 1.0]] == df.as_array(["b", "a"])

    df = DF([[np.float64(1.0), 1.1]], "a:double,b:int")
    assert [[1.0, 1]] == df.as_array()
    assert isinstance(df.as_array()[0][0], float)
    assert isinstance(df.as_array()[0][1], int)

    df = DF([[pd.Timestamp("2020-01-01"), 1.1]], "a:datetime,b:int")
    df.native["a"] = pd.to_datetime(df.native["a"])
    assert [[datetime(2020, 1, 1), 1]] == df.as_array()
    assert isinstance(df.as_array()[0][0], datetime)
    assert isinstance(df.as_array()[0][1], int)

    df = DF([[pd.NaT, 1.1]], "a:datetime,b:int")
    df.native["a"] = pd.to_datetime(df.native["a"])
    assert isinstance(df.as_array()[0][0], datetime)
    assert isinstance(df.as_array()[0][1], int)

    df = DF([[1.0, 1.1]], "a:double,b:int")
    assert [[1.0, 1]] == df.as_array(type_safe=True)
    assert isinstance(df.as_array()[0][0], float)
    assert isinstance(df.as_array()[0][1], int)


def test_nested():
    data = [[[json.dumps(dict(b=[30, "40"]))]]]
    s = expression_to_schema("a:[{a:str,b:[int]}]")
    df = DF(data, "a:[{a:str,b:[int]}]")
    a = df.as_array(s, type_safe=True)
    assert [[[dict(a=None, b=[30, 40])]]] == a

    data = [[json.dumps(["1", 2])]]
    s = expression_to_schema("a:[int]")
    df = DF(data, "a:[int]")
    a = df.as_array(s, type_safe=True)
    assert [[[1, 2]]] == a


def test_nan_none():
    df = DF([[None, None]], "b:str,c:double", True)
    assert df.native.iloc[0, 0] is None
    arr = df.as_array(null_safe=True)[0]
    assert arr[0] is None
    assert math.isnan(arr[1])

    df = DF([[None, None]], "b:int,c:bool", True)
    arr = df.as_array(type_safe=True)[0]
    assert np.isnan(arr[0])  # TODO: this will cause inconsistent behavior cross engine
    assert np.isnan(arr[1])  # TODO: this will cause inconsistent behavior cross engine

    df = DF([["a", 1.1], [None, None]], "b:str,c:double", True)
    arr = df.as_array()[1]
    assert arr[0] is None
    assert math.isnan(arr[1])

    df = DF([], "b:str,c:double", True)
    assert len(df.as_array()) == 0


def test_safe_group_by_apply():
    df = DF([["a", 1], ["a", 2], [None, 3]], "b:str,c:long", True)

    def _m1(df):
        _ensure_compatible_index(df)
        df["ct"] = df.shape[0]
        return df

    res = safe_groupby_apply(df.native, ["b"], _m1)
    _ensure_compatible_index(res)
    assert 3 == res.shape[0]
    assert 3 == res.shape[1]
    assert [["a", 1, 2], ["a", 2, 2], [None, 3, 1]] == res.values.tolist()

    res = safe_groupby_apply(df.native, [], _m1)
    _ensure_compatible_index(res)
    assert 3 == res.shape[0]
    assert 3 == res.shape[1]
    assert [["a", 1, 3], ["a", 2, 3], [None, 3, 3]] == res.values.tolist()

    df = DF([[1.0, "a"], [1.0, "b"], [None, "c"], [None, "d"]], "b:double,c:str", True)
    res = safe_groupby_apply(df.native, ["b"], _m1)
    assert [[1.0, "a", 2], [1.0, "b", 2],
            [float('nan'), "c", 2], [float('nan'), "d", 2]].__repr__() == res.values.tolist().__repr__()


class DF(object):  # This is a mock
    def __init__(self, data, schema, enforce=False):
        s = expression_to_schema(schema)
        df = pd.DataFrame(data, columns=s.names)
        self.native = enforce_type(df, s, enforce)

    def as_array(self, cols=None, type_safe=False, null_safe=False):
        if cols is None or isinstance(cols, pa.Schema):
            return list(as_array_iterable(
                self.native, schema=cols, type_safe=type_safe, null_safe=null_safe))
        if isinstance(cols, list):
            os = to_schema(self.native)
            s = pa.schema([os.field(x) for x in cols])
            return list(as_array_iterable(
                self.native, schema=s, type_safe=type_safe, null_safe=null_safe))



import json
from datetime import datetime

import numpy as np
import pandas as pd
from pytest import raises
from triad.collections.dataframe import PandasDataFrame
from triad.collections.schema import Schema, SchemaError
from triad.exceptions import InvalidOperationError, NoneArgumentError


def test_init():
    raises(SchemaError, lambda: PandasDataFrame())
    raises(SchemaError, lambda: PandasDataFrame(schema=Schema()))

    df = PandasDataFrame(schema="a:str,b:int")
    assert df.count() == 0
    assert df.schema == "a:str,b:int"
    assert Schema(df.native) == "a:str,b:int"

    pdf = pd.DataFrame([["a", 1], ["b", 2]])
    raises(SchemaError, lambda: PandasDataFrame(pdf))
    df = PandasDataFrame(pdf, "a:str,b:str")
    assert [["a", "1"], ["b", "2"]] == df.native.values.tolist()
    df = PandasDataFrame(pdf, "a:str,b:int")
    assert [["a", 1], ["b", 2]] == df.native.values.tolist()
    df = PandasDataFrame(pdf, "a:str,b:double")
    assert [["a", 1.0], ["b", 2.0]] == df.native.values.tolist()

    pdf = pd.DataFrame([["a", 1], ["b", 2]], columns=["a", "b"])["b"]
    assert isinstance(pdf, pd.Series)
    df = PandasDataFrame(pdf, "b:str")
    assert [["1"], ["2"]] == df.native.values.tolist()
    df = PandasDataFrame(pdf, "b:double")
    assert [[1.0], [2.0]] == df.native.values.tolist()

    pdf = pd.DataFrame([["a", 1], ["b", 2]], columns=["x", "y"])
    df = PandasDataFrame(pdf)
    assert df.schema == "x:str,y:long"
    df = PandasDataFrame(pdf, "y:str,x:str")
    assert [["1", "a"], ["2", "b"]] == df.native.values.tolist()
    ddf = PandasDataFrame(df)
    assert [["1", "a"], ["2", "b"]] == ddf.native.values.tolist()
    assert df.native is ddf.native  # no real copy happened

    df = PandasDataFrame([["a", 1], ["b", "2"]], "x:str,y:double")
    assert [["a", 1.0], ["b", 2.0]] == df.native.values.tolist()

    df = PandasDataFrame([], "x:str,y:double")
    assert [] == df.native.values.tolist()

    raises(ValueError, lambda: PandasDataFrame(123))


def test_simple_methods():
    df = PandasDataFrame([], "a:str,b:int")
    assert df.as_pandas() is df.native
    assert df.empty()
    assert 0 == df.count()
    raises(IndexError, lambda: df.peek_array())
    raises(IndexError, lambda: df.peek_dict())
    assert df.is_local

    df = PandasDataFrame([["a", 1], ["b", "2"]], "x:str,y:double")
    assert df.as_pandas() is df.native
    assert not df.empty()
    assert 2 == df.count()
    assert ["a", 1.0] == df.peek_array()
    assert dict(x="a", y=1.0) == df.peek_dict()


def test_nested():
    #data = [[dict(a=1, b=[3, 4], d=1.0)], [json.dumps(dict(b=[30, "40"]))]]
    #df = PandasDataFrame(data, "a:{a:str,b:[int]}")
    #a = df.as_array(type_safe=True)
    #assert [[dict(a="1", b=[3, 4])], [dict(a=None, b=[30, 40])]] == a

    data = [[[json.dumps(dict(b=[30, "40"]))]]]
    df = PandasDataFrame(data, "a:[{a:str,b:[int]}]")
    a = df.as_array(type_safe=True)
    assert [[[dict(a=None, b=[30, 40])]]] == a


def test_drop():
    df = PandasDataFrame([], "a:str,b:int").drop(["a"])
    assert df.empty()
    assert df.schema == "b:int"
    raises(InvalidOperationError, lambda: df.drop(["b"]))  # can't be empty
    raises(InvalidOperationError, lambda: df.drop(["x"]))  # cols must exist

    df = PandasDataFrame([["a", 1]], "a:str,b:int").drop(["a"])
    assert df.schema == "b:int"
    raises(InvalidOperationError, lambda: df.drop(["b"]))  # can't be empty
    raises(InvalidOperationError, lambda: df.drop(["x"]))  # cols must exist
    assert [[1]] == df.as_pandas().values.tolist()


def test_as_array():
    df = PandasDataFrame([], "a:str,b:int")
    assert [] == df.as_array()
    assert [] == df.as_array(type_safe=True)
    assert [] == list(df.as_array_iterable())
    assert [] == list(df.as_array_iterable(type_safe=True))

    df = PandasDataFrame([["a", 1]], "a:str,b:int")
    assert [["a", 1]] == df.as_array()
    assert [["a", 1]] == df.as_array(["a", "b"])
    assert [[1, "a"]] == df.as_array(["b", "a"])

    # prevent pandas auto type casting
    df = PandasDataFrame([[1.0, 1.1]], "a:double,b:int")
    assert [[1.0, 1]] == df.as_array()
    assert isinstance(df.as_array()[0][0], float)
    assert isinstance(df.as_array()[0][1], int)
    assert [[1.0, 1]] == df.as_array(["a", "b"])
    assert [[1, 1.0]] == df.as_array(["b", "a"])

    df = PandasDataFrame([[np.float64(1.0), 1.1]], "a:double,b:int")
    assert [[1.0, 1]] == df.as_array()
    assert isinstance(df.as_array()[0][0], float)
    assert isinstance(df.as_array()[0][1], int)

    df = PandasDataFrame([[pd.Timestamp("2020-01-01"), 1.1]], "a:datetime,b:int")
    df.native["a"] = pd.to_datetime(df.native["a"])
    assert [[datetime(2020, 1, 1), 1]] == df.as_array()
    assert isinstance(df.as_array()[0][0], datetime)
    assert isinstance(df.as_array()[0][1], int)

    df = PandasDataFrame([[pd.NaT, 1.1]], "a:datetime,b:int")
    df.native["a"] = pd.to_datetime(df.native["a"])
    assert isinstance(df.as_array()[0][0], datetime)
    assert isinstance(df.as_array()[0][1], int)

    df = PandasDataFrame([[1.0, 1.1]], "a:double,b:int")
    assert [[1.0, 1]] == df.as_array(type_safe=True)
    assert isinstance(df.as_array()[0][0], float)
    assert isinstance(df.as_array()[0][1], int)


def test_as_dict_iterable():
    df = PandasDataFrame([[pd.NaT, 1.1]], "a:datetime,b:int")
    assert [dict(a=pd.NaT, b=1)] == list(df.as_dict_iterable())
    df = PandasDataFrame([["2020-01-01", 1.1]], "a:datetime,b:int")
    assert [dict(a=datetime(2020, 1, 1), b=1)] == list(df.as_dict_iterable())


def _test_as_array_perf():
    s = Schema()
    arr = []
    for i in range(100):
        s.append(f"a{i}:int")
        arr.append(i)
    for i in range(100):
        s.append(f"b{i}:int")
        arr.append(float(i))
    for i in range(100):
        s.append(f"c{i}:str")
        arr.append(str(i))
    data = []
    for i in range(5000):
        data.append(list(arr))
    df = PandasDataFrame(data, s)
    res = df.as_array()
    res = df.as_array(type_safe=True)
    nts, ts = 0.0, 0.0
    for i in range(10):
        t = datetime.now()
        res = df.as_array()
        nts += (datetime.now() - t).total_seconds()
        t = datetime.now()
        res = df.as_array(type_safe=True)
        ts += (datetime.now() - t).total_seconds()
    print(nts, ts)



from typing import Any, Iterable, List, Optional

from triad.collections.dataframe.dataframe import (
    DataFrame,
    LocalDataFrame,
    _get_schema_change,
)
from triad.collections.schema import Schema
from triad.exceptions import InvalidOperationError
from triad.utils.iter import EmptyAwareIterable, make_empty_aware
from triad.utils.pyarrow import apply_schema


class IterableDataFrame(LocalDataFrame):
    def __init__(  # noqa: C901
        self, df: Any = None, schema: Any = None, metadata: Any = None
    ):
        if df is None:
            idf: Iterable[Any] = []
            orig_schema: Optional[Schema] = None
        elif isinstance(df, IterableDataFrame):
            idf = df.native
            orig_schema = df.schema
        elif isinstance(df, DataFrame):
            idf = df.as_array_iterable(type_safe=False)
            orig_schema = df.schema
        elif isinstance(df, Iterable):
            idf = df
            orig_schema = None
        else:
            raise ValueError(f"{df} is incompatible with IterableDataFrame")
        schema, pos = _get_schema_change(orig_schema, schema)
        super().__init__(schema, metadata)
        self._pos = pos
        self._native = make_empty_aware(self._preprocess(idf))

    @property
    def native(self) -> EmptyAwareIterable[Any]:
        return self._native

    def empty(self) -> bool:
        return self.native.empty

    def peek_array(self) -> Any:
        return list(self.native.peek())

    def count(self, persist: bool = False) -> int:
        raise InvalidOperationError("Impossible to count an IterableDataFrame")

    def drop(self, cols: List[str]) -> DataFrame:
        try:
            schema = self.schema - cols
        except Exception as e:
            raise InvalidOperationError(str(e))
        if len(schema) == 0:
            raise InvalidOperationError("Can't remove all columns of a dataframe")
        return IterableDataFrame(self, schema)

    def as_array(
        self, columns: Optional[List[str]] = None, type_safe: bool = False
    ) -> List[Any]:
        return list(self.as_array_iterable(columns, type_safe=type_safe))

    def as_array_iterable(
        self, columns: Optional[List[str]] = None, type_safe: bool = False
    ) -> Iterable[Any]:
        if columns is None:
            if not type_safe:
                for item in self.native:
                    yield item
            else:
                for item in apply_schema(
                    self.schema.pa_schema,
                    self.native,
                    copy=True,
                    deep=True,
                    str_as_json=True,
                ):
                    yield item
        else:
            df = IterableDataFrame(self, self.schema.extract(columns))
            for item in df.as_array_iterable(type_safe=type_safe):
                yield item

    def _preprocess(self, raw_data: Iterable[Any]) -> Iterable[Any]:
        if len(self._pos) == 0:
            for x in raw_data:
                yield x
        else:
            for x in raw_data:
                yield [x[i] for i in self._pos]



from pytest import raises
from triad.utils.function import (extract_function_arg_names,
                                  extract_function_io_types, safe_invoke)


def test__extract_func_id():
    assert ([], None) == extract_function_io_types(efi0)
    assert ([("a", int), ("b", None)], None) == extract_function_io_types(efi1)
    assert ([("a", None)], None) == extract_function_io_types(efi2)
    assert ([("a", None), ("b", int), ("arg", None), ("karg", None)],
            int) == extract_function_io_types(efi3)
    assert ([("a", int), ("b", None)], None) == extract_function_io_types(efi4, True)
    raises(TypeError, lambda: extract_function_io_types(efi4))
    raises(TypeError, lambda: extract_function_io_types(efi4, False))
    assert ([("a", int)], None) == extract_function_io_types(efi5, True)
    raises(TypeError, lambda: extract_function_io_types(efi5))
    raises(TypeError, lambda: extract_function_io_types(efi5, False))


def test__extract_function_arg_names():
    assert [] == extract_function_arg_names(si_0)
    assert ["a", "b"] == extract_function_arg_names(si_1)
    assert ["*a", "**b"] == extract_function_arg_names(si_2)
    assert ["x", "y", "*a", "**b"] == extract_function_arg_names(si_3)
    assert ["x", "y"] == extract_function_arg_names(si_4)
    assert ["x", "y", "*a", "**b"] == extract_function_arg_names(si_5)


def test__safe_invoke():
    assert safe_invoke("si_0") == 1
    assert safe_invoke("si_0", 1, 2, 3, a=1) == 1
    assert safe_invoke("si_1", 1, 2, 5) == 3
    assert safe_invoke("si_1", 1, 2, a=1) == 2
    assert safe_invoke("si_1", 1, 2, 3, b=3, c="d") == 4
    assert safe_invoke("si_1", b=3, a=2) == 5
    assert safe_invoke("si_2", 3, 2) == 5
    assert safe_invoke("si_2", b=3, a=2) == 5
    assert safe_invoke("si_2", 1, 2, 3, b=3, a=2) == 11
    assert safe_invoke("si_3", 1, 2, 3, b=3, a=2) == 10
    assert safe_invoke("si_3", 1, 2, 3, b=3, x=10) == 18
    assert safe_invoke("si_3", y=3, x=10) == 30
    assert safe_invoke("si_4", 6) == 3
    assert safe_invoke("si_4", 6, 3) == 2
    assert safe_invoke("si_4", 10, y=2) == 5
    assert safe_invoke("si_4", a=200, y=3, x=6) == 2
    assert safe_invoke("si_4", 6, x=12) == 2
    assert safe_invoke("si_5", 6) == 60
    assert safe_invoke("si_5", 5, x=12, y=4) == 53
    assert safe_invoke("si_5", **{"x": 8}) == 80
    assert safe_invoke("si_5", 6, 3, 1, 2, a=10) == 31
    assert safe_invoke("si_5", 6, 3, a=10) == 28
    raises(TypeError, lambda: safe_invoke("si_5", a=10))


def si_0():
    return 1


def si_1(a, b):
    return a + b


def si_2(*a, **b):
    return sum(a) + sum(b.values())


def si_3(x, y, *a, **b):
    return x * y + sum(a) + sum(b.values())


def si_4(x, y=2):
    return x / y


def si_5(x, y=10, *a, **b):
    return x * y + sum(a) + sum(b.values())


def efi0():
    pass


def efi1(a: int, b):
    pass


def efi2(a) -> None:
    pass


def efi3(a, b: "int", *arg, **karg) -> "int":
    pass


def efi4(a: int, b: "dummy") -> None:
    pass


def efi5(a: int) -> "dummy":
    pass



import inspect
from typing import Any, Dict, List, Optional, Tuple
from triad.utils.convert import to_type, to_function


def extract_function_io_types(
    func: Any, ignore_unknown_type: bool = False
) -> Tuple[List[Tuple[str, Optional[type]]], Optional[type]]:
    """Inspect the function and return input output specs

    :param func: input function

    :return: input name, type tuples and output type (None if no return)
    """
    sig = inspect.signature(func)
    inputs: List[Tuple[str, Optional[type]]] = []
    for k, v in sig.parameters.items():
        ktype: Optional[type] = None
        try:
            t = to_type(v.annotation)
            if t != inspect._empty:  # type: ignore
                ktype = t
        except Exception:
            if not ignore_unknown_type:
                raise
        inputs.append((k, ktype))
    otype: Optional[type] = None
    try:
        t = to_type(sig.return_annotation)
        if t != inspect._empty and t != type(None):  # type: ignore # noqa E721
            otype = t
    except Exception:
        if not ignore_unknown_type:
            raise
    return inputs, otype


def extract_function_arg_names(func: Any) -> List[str]:
    """Extract function argument names

    :param func: the input function
    :return: a list of names
    """
    sig = inspect.signature(func)
    args: List[str] = []
    for k, v in sig.parameters.items():
        raw = v.__str__()
        name = k
        if raw.startswith("**"):
            name = "**" + k
        elif raw.startswith("*"):
            name = "*" + k
        args.append(name)
    return args


def safe_invoke(func: Any, *args: Any, **kwargs: Any) -> Any:
    """Safely invoke a function when the given arguments is a superset
    of what can be accepted

    :param func: the function
    :raises TypeError: the arguments don't match and can't be tolerated
    """
    f = to_function(func)
    sig = inspect.signature(f)
    ag: List[Any] = list(args)
    kag: Dict[str, Any] = dict(kwargs)
    a: List[Any] = []
    for _, v in sig.parameters.items():
        k = v.name
        if not v.__str__().startswith("*"):
            if k in kag:
                a.append(kag.pop(k))
            elif len(ag) > 0:
                a.append(ag.pop(0))
            elif v.default is not inspect.Parameter.empty:
                a.append(v.default)
            else:
                raise TypeError(f"Unable to set value for {k}")
        elif v.__str__().startswith("**"):
            return f(*a, *ag, **kag)
        else:
            a += ag
            ag = []
    return f(*a)



from triad.collections.dataframe import ArrayDataFrame, DataFrame
from triad.collections.schema import Schema


def test_show():
    df = ArrayDataFrame(schema="a:str,b:str")
    df.show()

    s = " ".join(["x"] * 2)
    df = ArrayDataFrame([[s, 1], ["b", 2]], "a:str,b:str")
    df.show()

    s = " ".join(["x"] * 200)
    df = ArrayDataFrame([[s, 1], ["b", 2]], "a:str,b:str")
    df.show()

    s = " ".join(["x"] * 200)
    df = ArrayDataFrame([[s, 1], ["b", s]], "a:str,b:str")
    df.show()

    s = "".join(["x"] * 2000)
    df = ArrayDataFrame([[s, 1], ["b", None]], "a:str,b:str")
    df.show()

    s = " ".join(["x"] * 20)
    schema = [f"a{x}:str" for x in range(20)]
    data = [[f"aasdfjasdfka;sdf{x}:str" for x in range(20)]]
    df = ArrayDataFrame(data, schema)
    df.show(best_width=2)

    s = " ".join(["x"] * 200)
    df = ArrayDataFrame([[s, 1], ["b", "s"]], "a:str,b:str", metadata=dict(a=1, b=2))
    df.show(n=1, show_count=True, title="abc")


def test_lazy_schema():
    df = MockDF([["a", 1], ["b", 2]], "a:str,b:str")
    assert callable(df._schema)
    assert df.schema == "a:str,b:str"


def test_get_info_str():
    df = ArrayDataFrame([["a", 1], ["b", 2]], "a:str,b:str")
    assert '{"schema": "a:str,b:str", "type": '
    '"tests.collections.dataframe.test_dataframe.MockDF", "metadata": {}}' == df.get_info_str()


class MockDF(ArrayDataFrame):
    def __init__(self, df=None, schema=None, metadata=None):
        super(). __init__(df=df, schema=schema, metadata=metadata)
        DataFrame.__init__(self, lambda: Schema(schema))
